from __future__ import annotations

from .general import default_print
from .global_control import disable
from .global_control import enable
from .global_control import setup
from .global_control import unload
from .logger import lk

__version__ = '5.1.5'
